#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include "btree.h"

int BTreeBlock_init(BTreeBlock* block) {
	block->isLeaf = false;
	block->numNodes = 0;
	block->nextBlock = NULL;
	block->prevBlock = NULL;

	memset(block->nodes, '\0', sizeof(BTreeNode*) * MAX_BLOCK_WIDTH);
}

int BTree_init(BTree* tree) {
	tree->head = malloc(sizeof(BTreeBlock));
	BTreeBlock_init(tree->head);
	tree->head->isLeaf = true;
}

int BTreeNode_init(BTreeNode* node, BTree* tree) {
	node->tree = tree;
	node->key = NULL;
	node->value = NULL;
	node->rchild = NULL;
	node->lchild = NULL;
}


bool BTreeBlock_check(BTreeBlock* block) {
	BTreeBlock* currentBlock = block;
	bool isOK = true;
	int i;
	BTreeNode* currentNode;
	BTreeNode* prevNode;
	for (i = 0; i < MAX_BLOCK_WIDTH; i++) {
		currentNode = currentBlock->nodes[i];
		if (currentNode == NULL) {
			break;
		}
		
		if (currentNode->lchild != NULL) {
			isOK = isOK && (
				strcmp(
					currentNode->key,
					currentNode->lchild->nodes[0]->key
				) > 0
			);
			if (isOK != true) {
				printf("failed at left child\n");
				printf("node: %s\n", currentNode->key);
				return isOK;
			}
			isOK = isOK && BTreeBlock_check(currentNode->lchild);
		}

		if (currentNode->rchild != NULL) {
			isOK = isOK && (
				strcmp(
					currentNode->key,
					currentNode->rchild->nodes[0]->key
				) <= 0
			);
			if (isOK != true) {
				printf("failed at right child\n");
				printf("node: %s\n", currentNode->key);
				return isOK;
			}
			isOK = isOK && BTreeBlock_check(currentNode->rchild);
		}

		if (i > 0) {
			isOK = isOK && (
				strcmp(
					currentNode->key,
					currentBlock->nodes[i-1]->key
				) >= 0
			);
			if (isOK != true) {
				printf("failed at sibling\n");
				printf("node: %s\n", currentNode->key);
				return isOK;
			}
		}
	}
	return isOK;
}
int BTreeBlock_print(BTreeBlock* block) {
	BTreeBlock* currentBlock = block;

	printf("[");
	int i;
	BTreeNode* currentNode;
	for (i = 0; i < MAX_BLOCK_WIDTH; i++) {
		currentNode = currentBlock->nodes[i];
		if (currentNode != NULL) {
			if (currentNode->value != NULL) {
				printf("%s:%d, ", currentNode->key, *(int*)currentNode->value);
			} else {
				printf("%s, ", currentNode->key);
			}
		} else {
			printf("NULL, ");
		}
	}
	printf("]\n");
	for (i = 0; i < MAX_BLOCK_WIDTH; i++) {
		currentNode = currentBlock->nodes[i];
		if (currentNode != NULL) {
			if (currentNode->lchild != NULL) {
				printf("%s l->", currentNode->key);
				BTreeBlock_print(currentNode->lchild);
			}
			if (currentNode->rchild != NULL) {
				printf("%s r->", currentNode->key);
				BTreeBlock_print(currentNode->rchild);
			}
		}
	}
}

void array_insert(void* base, int index, void* item, size_t nmemb, size_t size) {
	memmove(
		(char*)base + (index * size) + size,
	       	(char*)base + (index * size),
	       	(nmemb - index) * size
	);

	memcpy(
		(char*)base + (index*size),
		&item,
		size
	);
}

void array_remove(void* base, int index, size_t nmemb, size_t size) {
	memmove(
		(char*)base + (index * size),
	       	(char*)base + (index * size) + size,
	       	(nmemb - index) * size
	);
	memset(
		(char*)base + (nmemb -1) * size,
		'\0',
		size
	);
}

// Merge block 1 into block 2
int BTreeBlock_merge(BTreeBlock* block1, BTreeBlock* block2) {
	block2->prevBlock = block1->prevBlock;
	memmove(
		block2->nodes + block1->numNodes,
		block2->nodes,
		block2->numNodes * sizeof(BTreeNode*)
	);
	memcpy(
		block2->nodes,
		block1->nodes,
		block1->numNodes * sizeof(BTreeNode*)
	);
	// TODO: free these nodes
	memset(
		block1->nodes,
		'\0',
		sizeof(BTreeNode*) * block1->numNodes
	);
	block2->numNodes += block1->numNodes;
}

int BTreeBlock_steal(BTreeBlock* theftTarget, size_t index, BTreeBlock* block) {
	size_t newPosition = 0;
	BTreeNode* targetNode = theftTarget->nodes[index];
	
	if (index == 0) {
		newPosition = block->numNodes;
	}
	
	array_insert(
		block->nodes,
		newPosition,
		theftTarget->nodes[index],
		block->numNodes,
		sizeof(BTreeNode*)	
	);

	array_remove(
		theftTarget->nodes,
		index,
		theftTarget->numNodes,
		sizeof(BTreeNode*)
	);
	theftTarget->numNodes--;
}

int BTreeBlock_removeInternal(BTree* tree, BTreeSearchResult searchResult) {
	size_t parentIndex = searchResult.parentIndex;
	size_t index = searchResult.index;
	BTreeBlock* parentBlock = searchResult.parentBlock;
	BTreeNode* parentNode = searchResult.parentNode;
	BTreeBlock* block = searchResult.block;
	BTreeNode* node = block->nodes[searchResult.index];
	if (
		block == parentNode->lchild 
		&& parentNode->rchild->numNodes > MIN_BLOCK_WIDTH-1
	) {
		printf("Stealing from right sibling\n");
	} else if (
		parentNodeIndex < parentBlock->numNodes-1 
		&& parentNode->lchild->numNodes > MIN_BLOCK_WIDTH-1
	) {
		printf("Stealing from left sibling\n");
		BTreeNode* targetNode = parentNode->lchild[numNodes-1];
		targetNode->lchild = targetNode->rchild;
		targetNode->rchild = node->lchild;
		char* parentNodeKey = parentNode->key;
		
		BTreeBlock_steal(
			parentNode->lchild,
			parentNode->lchild->numNodes-1,
			block
		);
		parentNode->key = targetNode->key;
		block->nodes[0]->key = parentNodeKey;
	} else if {
		printf("Stealing from left sibling\n");
	} else {
		printf("Merging with someone\n");
	}
	
}

int BTreeBlock_remove(BTree* tree, BTreeSearchResult searchResult) {
	size_t parentIndex = searchResult.parentIndex;
	size_t index = searchResult.index;
	BTreeBlock* parentBlock = searchResult.parentBlock;
	BTreeNode* parentNode = searchResult.parentNode;
	BTreeBlock* block = searchResult.block;
	BTreeNode* node = block->nodes[searchResult.index];
	char* targetKey = node->key;

	BTreeBlock* theftTarget = NULL;
	size_t theftTargetIndex = 0;

	if (block->prevBlock->numNodes > MIN_BLOCK_WIDTH) {
		printf("Stealing from left sibling\n");
		theftTarget = block->prevBlock;
		theftTargetIndex = theftTarget->numNodes-1;
	} else if (block->nextBlock->numNodes > MIN_BLOCK_WIDTH) {
		printf("Stealing from right sibling\n");
		theftTarget = block->nextBlock;
		theftTargetIndex = 0;
	}

	if (theftTarget != NULL) {
		array_remove(
			block->nodes,
			index,
			block->numNodes,
			sizeof(BTreeNode*)
		);
		block->numNodes--;
		BTreeBlock_steal(theftTarget, theftTargetIndex, block);

	} else if (
		block->numNodes <= MIN_BLOCK_WIDTH
	){
		printf("Merging...\n");
		array_remove(
			block->nodes,
			index,
			block->numNodes,
			sizeof(BTreeNode*)
		);
		block->numNodes--;
		if (block == parentNode->lchild) {
			BTreeBlock_merge(block, parentNode->rchild);
			block = parentNode->rchild;
		} else {
			BTreeBlock_merge(parentNode->lchild, block);
			if (parentIndex > 0) {
				parentBlock->nodes[parentIndex -1]->rchild = block;
			}
		}
		array_remove(
			parentBlock->nodes,
			parentIndex,
			parentBlock->numNodes,
			sizeof(BTreeNode*)
		);
		free(parentNode);
		parentBlock->numNodes--;
	} else {
		printf("Simple\n");
		array_remove(
			block->nodes,
			index,
			block->numNodes,
			sizeof(BTreeNode*)
		);
		block->numNodes--;
	}

	if (index == 0) {
		printf("targetkey is: %s\n", targetKey);
		searchResult = BTreeBlock_search(tree->head, targetKey, 2);
		searchResult.block->nodes[searchResult.index]->key = block->nodes[0]->key;
		printf("key is: %s\n", searchResult.block->nodes[searchResult.index]->key);
	}

	if (node != NULL) {
		printf("key is: %s\n", node->key);
	}

	free(node);
}

int BTree_delKey(BTree* tree, char * targetKey) {
	BTreeSearchResult searchResult = BTreeBlock_search(tree->head, targetKey, 1);
	BTreeBlock_remove(tree, searchResult);
}

int BTreeBlock_split(BTreeBlock* block, BTreeNode* parentNode, BTreeBlock* parentBlock) {
	size_t newBlockWidth = (MAX_BLOCK_WIDTH / 2);
	size_t newSiblingWidth = (MAX_BLOCK_WIDTH - (MAX_BLOCK_WIDTH / 2));

	BTreeNode* midNode = block->nodes[newBlockWidth];

	// Create a sibling block to split the nodes into
	BTreeBlock* siblingBlock = malloc(sizeof(BTreeBlock));
	BTreeBlock_init(siblingBlock);
	siblingBlock->isLeaf = block->isLeaf;

	// Create a parent block if one does not exists
	if (parentBlock == NULL) {
		BTreeBlock* newParentBlock = malloc(sizeof(BTreeBlock));
		BTreeBlock_init(newParentBlock);
		parentBlock = newParentBlock;
	} else if (block == parentNode->lchild) {
		// If the block is a left child, it's sibling should
		// take it's position in the parent
		parentNode->lchild = siblingBlock;
	}

	// Create and initialize the new parent node
	BTreeNode* newParentNode = malloc(sizeof(BTreeNode));
	BTreeNode_init(newParentNode, block->nodes[0]->tree);
	newParentNode->key = midNode->key;
	newParentNode->lchild = block;
	newParentNode->rchild = siblingBlock;

	// Move ceil(m/2) nodes into the sibling block
	memcpy(
		siblingBlock->nodes, 
		block->nodes + newBlockWidth, 
		sizeof(BTreeNode*)*newSiblingWidth
	);
	memset(
		block->nodes + newBlockWidth,
		'\0',
		sizeof(BTreeNode*)*newSiblingWidth
	);

	siblingBlock->nextBlock = block->nextBlock;
	block->nextBlock = siblingBlock;
	siblingBlock->prevBlock = block;
	
	block->numNodes = newBlockWidth;
	siblingBlock->numNodes = newSiblingWidth;


	if (block == block->nodes[0]->tree->head) {
		block->nodes[0]->tree->head = parentBlock;
	}
	BTreeBlock_insert(parentBlock, newParentNode);
	// Delete the middle node if this block not a leaf
	if (block->isLeaf == false) {
		siblingBlock->nodes[1]->lchild = siblingBlock->nodes[0]->rchild;
		memcpy(
			siblingBlock->nodes, 
			siblingBlock->nodes + 1,
			sizeof(BTreeNode*)
		);
		memset(
			siblingBlock->nodes + 1,
			'\0',
			sizeof(BTreeNode*)
		);
		siblingBlock->numNodes--;
	}
}

// This is terrible, replace if possible, (polymorphism in BTreeBlock_search?)
BTreeSearchResult BTreeBlock_getParent(BTree* tree, BTreeBlock* targetBlock) {
	size_t i;
	size_t targetKeyIndex;
	BTreeSearchResult result;
	result.block = NULL;
	result.parentBlock = NULL;
	result.parentNode = NULL;
	BTreeBlock* currentBlock = tree->head;
	while (1 == 1) {
		targetKeyIndex = 0;
		// TODO: replace this with binary search
		
		if (strcmp(currentBlock->nodes[0]->key, targetBlock->nodes[0]->key) > 0) {
			if (targetBlock != currentBlock->nodes[0]->lchild) {
				currentBlock = currentBlock->nodes[0]->lchild;
			} else {
				break;
			}
		} else {
			for (i = 0; i < currentBlock->numNodes; i++) {
				if (currentBlock->nodes[i] != NULL) {
					if(strcmp(currentBlock->nodes[i]->key, targetBlock->nodes[0]->key) <= 0) {
						targetKeyIndex = i;
					}
				} else {
					break;
				}
			}
			if (
				targetBlock != currentBlock->nodes[targetKeyIndex]->rchild
			) {
				if (i == 0) {
					break;
				} else {
					currentBlock = currentBlock->nodes[targetKeyIndex]->rchild;
				}
			} else {
				break;
			}
		}
	}
	result.parentBlock = currentBlock;
	result.parentNode = currentBlock->nodes[targetKeyIndex];
	result.parentIndex = targetKeyIndex;
	return result;
}

//TODO: refactor this to allow us to find the parent of the block when fullSearch = false
BTreeSearchResult BTreeBlock_search(BTreeBlock* start, char* targetKey, int searchMode) {
	// NOTE: if the value is not in the tree, this function will return the
	// index where the value WOULD be stored
	size_t i;
	size_t targetKeyIndex;
	BTreeSearchResult result;
	result.parentBlock = NULL;
	result.parentNode = NULL;
	result.parentIndex = 0;
	BTreeBlock* currentBlock = start;
	while (1 == 1) {
		targetKeyIndex = 0;

		if (strcmp(targetKey, currentBlock->nodes[0]->key) < 0) {
				
			if (
				currentBlock->isLeaf == false 

				&& searchMode == 1
			) {
				result.parentIndex = targetKeyIndex;
				result.parentBlock = currentBlock;
				result.parentNode = currentBlock->nodes[targetKeyIndex];
				currentBlock = result.parentNode->lchild;
				continue;
			} else {
				if (currentBlock->numNodes != 0 && currentBlock != currentBlock->nodes[0]->tree->head) {
					result = BTreeBlock_getParent(currentBlock->nodes[0]->tree,currentBlock);
				}
				result.block = currentBlock;
				result.index = targetKeyIndex;
				break;
			}
		}

		// TODO: replace this with binary search
		for (i = 0; i < currentBlock->numNodes; i++) {
			if (currentBlock->nodes[i] != NULL) {
				if(strcmp(targetKey, currentBlock->nodes[i]->key) >= 0) {
					targetKeyIndex = i;
				}
			}
		}
		if (
			searchMode == 2
			&& *currentBlock->nodes[targetKeyIndex]->key == *targetKey
		) {
			result.block = currentBlock;
			result.index = targetKeyIndex;
			break;
		} else if (
			currentBlock->isLeaf == false 
			&& searchMode != 0
		) {
			result.parentBlock = currentBlock;
			result.parentNode = currentBlock->nodes[targetKeyIndex];
			currentBlock = result.parentNode->rchild;
		} else {
			if (currentBlock->numNodes != 0 && currentBlock != currentBlock->nodes[0]->tree->head) {
				result = BTreeBlock_getParent(currentBlock->nodes[0]->tree,currentBlock);
			}

			if (searchMode == 2) {
				result.block = NULL;
			} else {
				result.block = currentBlock;
				result.index = targetKeyIndex;
			}
			break;
		}
	}

	return result;
}

int BTreeBlock_insert(BTreeBlock* block, BTreeNode* newNode) {
	BTreeSearchResult searchResult;
	size_t targetKeyIndex;

	// Special case for an empty root
	if (block->numNodes == 0) {
		block->nodes[0] = newNode;
		block->numNodes++;
		return 0;
	}

	// If the node value is empty then this is not a leaf
	// node, it was inserted into the parent currentBlock by a
	// split operation, keep it there.
	if (newNode->value != NULL) {
		searchResult = BTreeBlock_search(block, newNode->key, 1);
	} else {
		searchResult = BTreeBlock_search(block, newNode->key, 0);
	}

	block = searchResult.block;
	if(strcmp(newNode->key, block->nodes[0]->key) < 0) {
		targetKeyIndex = searchResult.index;
	} else {
		targetKeyIndex = searchResult.index+1;
	}

	array_insert(
		block->nodes,
		targetKeyIndex,
		newNode,
		block->numNodes,
		sizeof(BTreeNode*)
	);
	block->numNodes++;

	if (block->numNodes == MAX_BLOCK_WIDTH) {
		BTreeBlock_split(block, searchResult.parentNode, searchResult.parentBlock);
	}
}

int BTree_insert(BTree* tree, char* key, void* value) {
	BTreeNode* newNode = malloc(sizeof(BTreeNode));
	BTreeNode_init(newNode, tree);
	newNode->key = malloc(strlen(key)+1);
	strcpy(newNode->key, key);
	newNode->value = value;

	BTreeBlock_insert(tree->head, newNode);

	return 0;
}

int main() {
	BTree* myTree = malloc(sizeof(BTree));
	BTree_init(myTree);
	bool isOK;
	int* num = malloc(sizeof(int));
	*num = 1;

	int i;
	char input[100];
	for (i=0; i< 26; i++) {
		sprintf(input, "%c", i+65);
		//sprintf(input, "%d", i);
		BTree_insert(myTree, input, (void*)num);
	}
	isOK = BTreeBlock_check(myTree->head);
	if (isOK == true) {
		printf("Tree is ok.\n");
	} else {
		printf("Tree is not ok.\n");
	}
	printf("Deleting H...\n");
	BTree_delKey(myTree, "H");
	/*
	printf("Deleting G...\n");
	BTree_delKey(myTree, "G");
	*/
	printf("Deleting K...\n");
	BTree_delKey(myTree, "K");
	BTreeBlock_print(myTree->head);

	/*
	while (1==1) {
		char input[100];
		scanf("%s", input);
		printf("inserting %s.\n" , input);
		BTree_insert(myTree, input, (void*)num);
		BTreeBlock_print(myTree->head);
		isOK = BTreeBlock_check(myTree->head);
		if (isOK == true) {
			printf("Tree is ok.\n");
		} else {
			printf("Tree is not ok.\n");
		}
	}
	*/
		
	/*
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	printf("splitting block for key %s\n", newNode->key);
	BTree_insert(myTree, key3, (void*)num);
	BTree_insert(myTree, key, (void*)num);
	BTree_insert(myTree, key2, (void*)num);
	BTree_insert(myTree, key4, (void*)num);
	BTree_insert(myTree, key5, (void*)num);
	BTree_insert(myTree, key6, (void*)num);
	BTreeBlock_print(myTree->head);
	free(myTree);
	*/
}
